# rtftclx.tcl
# Tcl part of the package for rt_fifos
# after any change in this file, execute: mkPkg.tcl
#  

# Charles Andr
# April 1, 2000
# modified: June 25, 2000  (namespace)
# modified : June 29, 2000

package provide Rtftclx 1.3

##############################################

# list of procs

# RTF from RTLinux to Linux:
#
#      (exported) proc Rtf::openEmitter { rtfNum hdler } 
#
#      proc Rtf::closeEmitter { rtf } 
#
# initialization of the RtfTcl package
#
#      (exported) proc Rtf::Init { cfgFile } 
#
# Access to the data base for exchanged variables
# 
#      proc Rtf::varFullName {tok}
#
#      proc Rtf::varInternName {tok}
#
#      proc Rtf::varDisplayName {tok}
#
#      proc Rtf::varType {tok}
#
#      proc Rtf::varChannel {tok}
#
#      proc Rtf::varToDisplay {tok} 
#
#      (exported) proc varSetAction {tok value}
#
# Facilities 
#
#      proc Rtf::sizeOfType {type}
#
#      proc Rtf::isAToken {tok}
#
#      proc Rtf::isAnAction {a}
#
#      proc Rtf::id2token {funcName varName}
#
#      proc Rtf::registerVar { funcName varName type rtf action }
#
#      proc Rtf::registerRTF { fifo num dir hdl}
#      
#      (exported) proc Rtf::buildVarPanel { parent }
#
#       proc Rtf::configDisplay { }
#

# Communication support
#
#      (exported) proc Rtf::traceReader {}
#
#      (exported) proc Rtf::paramWriter { rtf token fn}
#
#      (exported) proc Rtf::MonWriter {}
#
#      (exported) proc Rtf::MonReader {}
#
#      (exported) proc Rtf::sendCmd {cmd}
#
#      (exported) proc Rtf::sendStop {}
#
#      (exported) proc Rtf::Quit {}
#
##############################################"


namespace eval Rtf { 
    variable internSize
    # CAN BE CUSTOMIZED BY THE USER (begin)
    # add here new types
    # currently supported types:
    #  (see Tcl's Binary Conversion types for further details)
    #   `c'  an 8-bit character code
    #   `a1' a character string of length 1
    #   `i'  a 32-bit integer little-endian buyte order 
    #   `s'  a 16-bit integer little-endian buyte order 
    array set internSize [list c 1 i 4 a1 1 s 2]
    # CAN BE CUSTOMIZED BY THE USER (end)
}    
    


namespace eval Rtf {
    variable numberOfTokens 0
    variable in2token
    variable token2var
    variable lastToken
    variable status        -1 ;#  -1 = read a token
    variable wgtPath
    variable command 
    variable response 
    variable emitterDB
    variable RTFdesc
    variable Hdl2RTF

    namespace export openEmitter Init buildVarPanel configDisplay
    namespace export sendCmd  sendStop Quit varSetAction
    namespace export traceReader MonReader paramWriter MonWriter

    puts stderr "procs exported from Rtf: [namespace export]"

    ##############################################
    # Procs dealing with RTF written by the RT program and read by a Linux program
    
    # neither create, nor delete fifos
    # deals only with the cat processus
    
    # args: rtfNum number (0..63)
    #       hdler      the call-back proc to call whenever the file is readable
    # returns the pipe id
    proc openEmitter { rtfNum hdler } {
	
	variable emitterDB
	
	# puts stderr "openEmitter $rtfNum $hdler"
	
	if { ($rtfNum <0) || ($rtfNum > 63) } {
	    puts stderr "Illegal rtf number: $rtfNum. Should be between 0 and 63."
	    return 2
	}
	set fn /dev/rtf$rtfNum
	if [info exists emitterDB($rtfNum)] {
	    puts stderr "$fn is already open."
	    return 2
	}
	
	if [catch [list open "| cat $fn" r ] rtf] {
	    puts stderr "Error when opening $fn for reading."
	    return 2
	}
	puts stderr "[exec ps u]"
	# get the description of the cat process
	set tmp [exec ps u | grep "cat $fn"]
	# replace sequences of multiple space characters with a single space
	regsub -all {\s+} $tmp " " tmp
	# get the pid (2nd field)
	set pid [lindex [split $tmp] 1]
	puts stderr "Pipe $rtf opened on $fn (pid=$pid)"
	lappend emitterDB(emitterList) $rtf
	set emitterDB(p,$rtf) $pid
	set emitterDB($fn) $rtf
	set emitterDB(h,$rtf) $hdler
	set emitterDB(r,$rtf) $fn
	set cmd [list fileevent $rtf readable $hdler]
	uplevel #0 $cmd
	# puts stderr "rtf=$rtf; cmd=$cmd"
	return $rtf
    }
    
    # args: rtf= name given to the RTF
    
    proc closeEmitter { rtf } {
	variable emitterDB
	
	if [info exists emitterDB(r,$rtf)] {
	    set pid $emitterDB(p,$rtf)
	    # kill the cat
	    exec kill -9 $pid
	    # close $rtf
	    set fn $emitterDB(r,$rtf)
	    puts stderr "closeEmitter $rtf ($fn)"
	    unset emitterDB($fn)
	    unset emitterDB(r,$rtf)
	    unset emitterDB(h,$rtf)
	} else {
	    puts stderr "$rtf is not opened as a rtfEmitter"
	}
	
    }
    

    # arg: configuration file
    proc Init { cfgFile } {
	variable numberOfTokens  
	variable in2token
	variable emitterDB
	variable RTFdesc
	variable Hdl2RTF
	
	global illegal applName rtTrace mon_snd mon_rcv
	
	set illegal 0
	set numberOfTokens 0

	# puts stderr "Entering Init"
	# get application-dependent information

	puts stderr "Sourcing $cfgFile"
	# source the configuration file
	uplevel #0 source $cfgFile
	
	# open RTFs
	
	foreach rtf [array names RTFdesc] {
	    set l  $RTFdesc($rtf)
	    set num [lindex $l 0]
	    set dir [lindex $l 1]
	    set hdl [lindex $l 2]

	    global $rtf 

	    switch -exact -- $dir {
		r {
		    set f [openEmitter $num $hdl]
		}
		w {
		    set f [rtf_open $num w]
		}
		# correct by construction: no default needed
	    }
	    eval set $rtf $f
	    # puts stderr "Init: set $rtf $f"
	}
	
	
	# build the dictionaries
	for {set k 0} {$k<$numberOfTokens} {incr k} {
	    eval set [varFullName $k] 0
	    set internName [varInternName $k]
	    # build reverse access to token
	    set in2token($internName) $k
	}
	
	# request for hide/show information
	configDisplay
	
    }
    
    
    ##############################################
    # auxiliary procs for Rtf access.
    
    
    # returns ${funcName}_${varName} associatedd with token $tok
    # assume token2var($tok) exists
    proc varFullName {tok} {
	variable token2var
	
	set l $token2var($tok)
	return "[lindex $l 0]_[lindex $l 1]"
    }
    
    # returns ${funcName},${varName} associatedd with token $tok
    # assume token2var($tok) exists
    proc varInternName {tok} {
	variable token2var
	
	set l $token2var($tok)
	return "[lindex $l 0],[lindex $l 1]"
    }
    
    # returns ${funcName}::${varName} associated with token $tok
    # assume token2var($tok) exists
    proc varDisplayName {tok} {
	variable token2var
	
	set l $token2var($tok)
	return "[lindex $l 0]\::[lindex $l 1]"
    }
    
    # returns the type of ${funcName}::${varName} associatedd with token $tok
    # assume token2type($tok) exists
    proc varType {tok} {
	variable token2var
	
	return [lindex $token2var($tok) 2]
    }
    
    
    proc varChannel {tok} {
	
	set chName [lindex $Rtf::token2var($tok) 3]
	# puts stderr "varChannel $tok -> $chName"
	upvar #0 $chName ch
	return $ch
    }
    
    # side effect: assign showOnDisplay($tok)
    proc varToDisplay {tok} {
	variable token2var
	global showOnDisplay
	
	set value [lindex $token2var($tok) 4]
	set showOnDisplay($tok) $value
	return $value
    }
    
    proc varSetAction {tok value} {

	set oldValue  [lindex $Rtf::token2var($tok) 4]
	if { "$oldValue" != "$value" } {
	    set Rtf::token2var($tok) [lreplace $Rtf::token2var($tok) 4 4 $value]
	}
    }
    
    # returns the size (in Bytes) of the argument's type
    # returns 0 if not defined
    proc sizeOfType {type} {
	variable internSize
	
	if [info exists internSize($type)] {
	    return $internSize($type)
	} else {
	    puts stderr "Type $type not supported."
	    return 0
	}
    }
    
    # returns 1 iff 0<= $tok < $numberOfTokens 
    proc isAToken {tok} {
	variable numberOfTokens
	
	if { $tok < 0 } {
	    return 0
	} 
	if { $tok < $numberOfTokens } {
	    return 1
	}
	return 0
    }
    
    # return 1 if action is r or w or 0
    #        0 otherwise
    proc isAnAction {a} {
	
	if { ( $a == "r" ) || ( $a == "w" ) || ( $a == 0 ) } { return 1 }
	return 0
    }
    
    # returns the token associated with the variable
    # returns -1 if not associated
    proc id2token {funcName varName} {
	variable in2token
	
	set internName  "$funcName,$varName"
	if [info exists in2token($internName)] {
	    return $in2token($internName)
	} else {
	    return -1
	}
    }
    
    # register a variable and associate a token
    
    # arguments:
    #    funcName ::=  {} | ident      // scope of the variable in the RT code
    #                                  // { } = global; ident = nom of a C function
    #    varName ::= ident             // name of the variable
    #        fullName ::=  ${funcName}_${varName}  // name given wrt Tcl
    #    type ::= ident                // defined above
    #    rtf ::= ident                 // name of the RTF that conveys the value
    #    action ::= r | w | 0          // r means that the tcl panel reads the variable
    #                                  // w means that the tcl panel writes the variable
    #                                  // 0 means that the variable must be hidden
    #
    #        token ::= integer
    
    # side effect: increments lastToken ans assigns token2var
    
    proc registerVar { funcName varName rtf type action } {
	variable token2var 
	variable numberOfTokens
	variable RTFdir
	
	puts stderr "Rtf::registerVar $funcName $varName $rtf $type $action"
	if [sizeOfType $type] {
	    # type is known
	    if [isAnAction $action] {
		# register
		set token2var($numberOfTokens) [list $funcName \
			$varName $type $rtf $action]
		incr numberOfTokens
		if [info exist RTFdir($rtf)] {
		    if { $RTFdir($rtf) != $action } {
			warningReport "RT-FIFO `$rtf' inconsistent use."
		    } else {
			set RTFdir($rtf) $action
		    }
		} else {
		    set RTFdir($rtf) $action
		}
	    } else {
		# illegal action
		puts stderr "$action is an illegal action. Ignore $funcName $varName."
	    }
	} else {
	    # unknown type: error + ignore
	    puts stderr "$type is an illegal type. Ignore $funcName $varName."
	}
    }

    proc registerRTF {fifo num dir hdl} {
	variable RTFdesc
	variable Hdl2RTF

	set RTFdesc($fifo) [list $num $dir $hdl]
	set Hdl2RTF($hdl) $fifo

    }
    
    
    #################################################################"
    
    
    proc traceReader {} {
	variable status 
	global illegal

	set rtfId $Rtf::Hdl2RTF(Rtf::traceReader)
	upvar #0 $rtfId rtf
	# puts stderr "MonWriter: rtfId:$rtfId; rtf=$rtf"

	
	
	if { $status == -1 } {
	    # read a token
	    if [catch {read $rtf 4} buf] {
		puts stderr "Error while reading $rtfId (token): $buf"
		# close $rtf
		Quit
	    }	
	    binary scan $buf "i" status
	    if [isAToken $status] {
		set illegal 0
	    } else {
		# illegal
		set illegal 1
		puts stderr "Illegal token:` $status'.  Ignored."
		set status -1
	    }
	} else {
	    # read $token2var($status)
	    # since $status is a legal token, all following attributes
	    # should be available
	    set var  [varFullName $status]
	    set type [varType $status]
	    set size [sizeOfType $type]
	    if [catch {read $rtf $size} buf] {
		puts stderr "Error while reading $rtfId ($var): $buf"
		close $rtf
		Quit
	    }	
	    global $var
	    binary scan $buf "$type" $var
	    set status -1
	    set illegal 0
	}
    }
    
    
    # write a tokenized variable
    proc paramWriter { rtf token fn} {
	
	
	set type [varType $token]
	set size [sizeOfType $type]
	upvar #0 $fn var
	set value $var
	# puts stderr "Rtf::paramWriter $rtf,$token,$fn -> type=$type,value=$value)"
	# first, send the token
	rtf_write $rtf [binary format "i" $token] 
	
	# then, send the value
	rtf_write $rtf [binary format "$type" $value] 
	
    }
    
    # send a command to the monitor, via rtf0
    # simplified format: a single character
    proc MonWriter {} {

	set rtfId $Rtf::Hdl2RTF(Rtf::MonWriter)
	upvar #0 $rtfId rtf
	# puts stderr "MonWriter: rtfId:$rtfId; rtf=$rtf"
	
	set firstChar [string index $Rtf::command 0]
	set written [rtf_write $rtf [binary format "a1" $firstChar ]]
	# puts stderr "Rtf::MonWriter `$firstChar' ($written bytes written)"
    }
    
    # receive a command from the monitor, via rtf1
    # simplified format: a single character
    proc MonReader {} {

	set rtfId $Rtf::Hdl2RTF(Rtf::MonReader)
	upvar #0 $rtfId rtf
	# puts stderr "MonReader: rtfId:$rtfId; rtf=$rtf"
	
	if [catch {read $rtf 1} buf] {
	    puts stderr "Error while reading $rtfId: $buf"
	    # catch {close $mon_rcv}
	    Quit
	}
	binary scan $buf "a1" Rtf::response
	if { $Rtf::response == "?" } {
	    puts stderr "RT application received an unexpected command."
	}
    }
    
    
    proc sendCmd {cmd} {

	
	set Rtf::command "$cmd"
	MonWriter
	vwait Rtf::response
    }
    
    
    proc sendStop {} {
	
	set Rtf::command "s"
	MonWriter
	vwait Rtf::response
	if { $Rtf::response == "q" } {
	    # it's time to kill the "cats"
	    puts stderr "Kill the cats on emitters."
	    
	    foreach rtf $Rtf::emitterDB(emitterList) {
		fileevent $rtf readable {}
		closeEmitter $rtf
	    }
	    
	    $Rtf::wgtPath(quit) configure -state normal
	}
    }
    
    proc Quit {} {
	global applName
	
	catch {exec rt_unload $applName}
	puts stderr "\n\t That's All Folks!\n"
	
	destroy .
    }
    
    
    ###########################################################
    
    # argument: parent window for the panel
    # returns the pack list for variables
    proc buildVarPanel { parent } {
	variable  numberOfTokens

	set frameList {}
	
	for {set k 0} {$k<$numberOfTokens} {incr k} {
	    
	    switch [varToDisplay $k] {
		0 {
		    # nothing to display
		}
		r {
		    set f [frame $parent.$k]
		    set l [label $f.l -text [varDisplayName $k] -width 12 -anchor e]
		    set fn [varFullName $k]
		    global $fn
		    set e [entry $f.e -textvariable $fn -width 10 -state disabled -relief flat]
		    pack $l $e -side left -padx 4
		    lappend frameList $f
		}
		w {
		    set f [frame $parent.$k]
		    set l [label $f.l -text [varDisplayName $k] -width 12 -anchor e]
		    set fn [varFullName $k]
		    global $fn
		    set e [entry $f.e -textvariable $fn -width 10]
		    pack $l $e -side left -padx 4
		    lappend frameList $f
		    set ch [varChannel $k]
		    bind $e <Return> [list Rtf::paramWriter $ch $k $fn]
		}
	    }
	}  
	return $frameList
    }


    # displays the list of exchange variables.
    # allows the user to hide variables
    # side effect: change action in token2var
    proc configDisplay { } {
	variable  numberOfTokens
	global showOnDisplay

	set parent [frame .mf -borderwidth 5 -relief groove]
	
	for {set k 0} {$k<$numberOfTokens} {incr k} {
	    
	    switch [varToDisplay $k] {
		0 {
		    # nothing to display
		}
		r {
		    set fn [varFullName $k]
		    
		    set bs [radiobutton $parent.${k}bs -text Show -command [list Rtf::varSetAction $k r] -variable showOnDisplay($k) -value r ]
		    set bh [radiobutton $parent.${k}bh -text Hide -command [list Rtf::varSetAction $k 0] -variable showOnDisplay($k) -value 0 ]
		    set l [label $parent.${k}l -text [varDisplayName $k] -width 30 -anchor e]
		    grid $bs $bh $l
		   
		}
		w {
		    set fn [varFullName $k]
		    
		    set bs [radiobutton $parent.${k}bs -text Show -command [list Rtf::varSetAction $k w] -variable showOnDisplay($k) -value w ]
		    set bh [radiobutton $parent.${k}bh -text Hide -command [list Rtf::varSetAction $k 0] -variable showOnDisplay($k) -value 0 ]
		    set l [label $parent.${k}l -text [varDisplayName $k] -width 30 -anchor e]
		    grid $bs $bh $l
	
		}
	    }
	} 
	
	set ba [button $parent.b -text Accept -command [list destroy .mf]]
	grid $ba  -column 0 -columnspan 3
	pack $parent
	tkwait window .mf
    }
}
