#!/bin/sh 
# Remark: launch wish wherever it is \
exec wish -f $0 $*
# genRtfTcl.tcl
# Charles Andr
# June 27, 2000
# modified: June 29, 2000

set versionNb 1.3
set generator RtfTcl
set LASTRELEASE "June 29, 2000"

set RTFTCL $env(RTFTCL)
set WDIR [pwd]
set LIBDIR $RTFTCL/lib
set BINDIR $RTFTCL/bin
set INCDIR $RTFTCL/include

proc isIdent {s} {

    set ss [string trim $s]
    return [regexp {[A-Za-z][A-Za-z0-9_]*} $ss]

}

proc isHandler {s} {

    set ss [string trim $s]
    if [regexp {([^:]*)::(.*)} $ss match arg1 arg2] {
	if { [isIdent $arg1] && [isIdent $arg2] } {
	    return 1
	} else {
	    return 0
	}
    } else {
	return [isIdent $ss]
    }
}

proc isFifoNum { s } {

    set ss [string trim $s]
    return [regexp {^[0-9]*$} $s]

}
proc isIn { vset elem } {

    upvar $vset set

    set pos [lsearch -exact $set $elem]
    if { $pos == -1 } { return 0 }
    return 1
}


# returns 1 if matches, 0 otherwise
proc parse {line keyword} {

    global fifo func var type num dir hdl

    switch -exact -- $keyword {
	RTFd {
	    set pat {.*rtf_dcl\(([^,]*),([^,]*),([^,]*),([^)]*)\).*}  
	    if [regexp $pat $line match fifo num dir hdl] {
		if [isIdent $fifo] {
		    if [isFifoNum $num] {
			switch -exact -- $dir {
			    r {
				# Linux -> RTLinux
				set dir w ;# Tcl interprets the reverse direction
			    }
			    w {
				# RTLinux -> Linux
				set dir r ;# Tcl interprets the reverse direction
			    }
			    default {
				errorReport "`$dir' is not a legal direction (should be r or w)."
			    }
			}
			if [isHandler $hdl] {
			    puts stderr "rtf_dcl= $fifo : $num,$dir,$hdl"
			    return 1
			} else {
			    errorReport "`$hdl' is an illegal Tcl proc identifier."
			}
		    } else {
			errorReport "`$num' is not a RTF number."
		    }
		} else {
		    errorReport "`$fifo' is not a legal C-function identifier."
		}
	    } else {
		# warningReport "parse `$line' doesn't match"
		return 0
	    }
	}
	putTo {
	    set pat {.*putTo\(([^,]*),([^,]*),([^,]*),([^)]*)\).*}    
	}
	caseDataFrom {
	    set pat {.*caseDataFrom\(([^,]*),([^,]*),([^,]*),([^)]*)\).*}    
	}
	default {
	    errorReport "`$keyword' illegal key word."
	}
    }
    # puts stderr "parse `$line' with pattern=`$pat'"
    if [regexp $pat $line match fifo func var type] {
	if [isIdent $fifo] {
	    if { ( $func == "" ) || [isIdent $func] || [string match {[ ]*} $func] } {
		if [isIdent $var] {
		    switch -- $type {
			Byte {
			    set t c
			}
			char {
			    set t a1
			}
			short {
			    set t s
			}
			int {
			    set t i
			}
			long {
			    set t i
			}
			default {
			    set t ""
			    warningReport "`$type': unsupported type.\nTaken as int."
			}
		    }
		    if { $t != "" } {
			puts stderr "$keyword $fifo $func $var $t"
			set type $t
			return 1
		    } else {
			set type i ;#default
			return 1
		    }
		} else {
		    warningReport "`$var' is not a legal C-variable identifier."
		    return  0
		}
	    } else {
		warningReport "`$func' is not a legal C-function identifier."
		return  0
	    }
	} else {
	    warningReport "`$fifo' is not a legal fifo identifier."
	    return  0
	}
    }
    # warningReport "parse `$line' doesn't match"
    return 0
}

proc errorReport { msg } {

    puts stderr "\n****** Error: $msg\nAbort\n\n"
    exit
}

proc warningReport { msg } {

    puts stderr "\n****** Warning: $msg\n"
 
}

# makes the header of the h-configure file
# and returns the file-id
proc headerHFile { } {

    global applName WDIR
    global versionNb generator LASTRELEASE

    # make a new configuration file name with a '.h' extension
    set filename ${applName}Cfg.h
    # check if output file can be created
    if [catch {open $WDIR/$filename w} fh] {
	errorReport "Cannot open $WDIR/$filenanme: $fh"
    }

    # let's write a little header
    puts $fh "/* $filename */"
    puts $fh "/* Creator: $generator v$versionNb ($LASTRELEASE) */"
    puts $fh "/* Copyright Charles Andr */"
    puts $fh "/*    Lab. I3S , NICE (F) email: andre@unice.fr */"
    puts $fh "/* Host:    [exec hostname] */"
    puts $fh "/* Date:    [exec date] */"
    puts $fh "/* Author:  [exec whoami] */"
    puts $fh "/* This configuration file is derived from file `$applName.c' */"
    puts $fh " "
    return $fh
}


# insert macro declarations in the h-cfg file
proc stuffHFile { fh } {

    global INCDIR
    puts $fh ""
    puts $fh "#include \"$INCDIR/coreH.h\""
    puts $fh ""

}

# makes the header of the tcl-configure file
# and returns the file-id
proc headerTclFile { } {

    global applName WDIR
    global versionNb generator LASTRELEASE

    # make a new configuration file name with a '.tcl' extension
    set filename ${applName}Cfg.tcl
    # check if output file can be created
    if [catch {open $WDIR/$filename w} ftcl] {
	errorReport "Cannot open $WDIR/$filenanme: $ftcl"
    }

    # let's write a little header
    puts $ftcl "# $filename"
    puts $ftcl "# Creator: $generator v$versionNb ($LASTRELEASE)"
    puts $ftcl "# Copyright Charles Andr "
    puts $ftcl "#    Lab. I3S , NICE (F) email: andre@unice.fr "
    puts $ftcl "# Host:    [exec hostname]"
    puts $ftcl "# Date:    [exec date]"
    puts $ftcl "# Author:  [exec whoami]"
    puts $ftcl "# This configuration file is derived from file `$applName.c'"
    puts $ftcl " "
    return $ftcl
}

set usage "usage: genRtfTcl  basename" 

if { $argc == 1 } {
    set applName [lindex $argv 0]
} else {
    puts stderr $usage
    exit
}

# make a void h-configuration file
set fh [headerHFile]
close $fh

set filtredProg [exec $BINDIR/extract $applName]

puts stderr "filtredProg=\n$filtredProg"


set signalSet {}
set tclList {}
set RTFList {}

foreach l [split $filtredProg '\n'] {
    

    if [parse $l putTo] {
	set fullName "${func}_${var}"
	if ![isIn signalSet $fullName] {
	    lappend signalSet $fullName
	    lappend tclList [list $fifo $func $var $type r]
	}
    } elseif [parse $l caseDataFrom] {
	set fullName "${func}_${var}"
	if ![isIn signalSet $fullName] {
	    lappend signalSet $fullName
	    lappend tclList [list $fifo $func $var $type w]
	}
    } elseif [parse $l RTFd] {
	if [info exists RTF2num($fifo)] {
	    warningReport "`$fifo' is already declared. \n New declaration ignored."
	} else {
	    if [info exists num2RTF($num)] {
		warningReport "`rtf$num' is already attached. \n New declaration ignored."
	    } else {
		set RTF2num($fifo) $num
		set num2RTF($num) $fifo
		lappend RTFList [list $fifo $num $dir $hdl]
	    }
	}
    }
}

# puts stderr "signalSet=$signalSet"
# puts stderr "tclList=$tclList"

# here generate configuration files

# generate ${applName}Cfg.h
set fh [headerHFile]

puts $fh "typedef enum \{"
foreach fn $signalSet {
    puts $fh "\t_tok_$fn,"
}
puts $fh "\t_tok_end\} _tokens;"
stuffHFile $fh
close $fh

# generate ${applName}Cfg.tcl
set ftcl [headerTclFile]

foreach l $RTFList {
    set fifo [lindex $l 0]
    set num  [lindex $l 1]
    set dir  [lindex $l 2]
    set hdl  [lindex $l 3]
    puts $ftcl "Rtf::registerRTF $fifo $num $dir $hdl"
}

foreach l $tclList {
    set fifo [lindex $l 0]
    set func [lindex $l 1]
    set var  [lindex $l 2]
    set type [lindex $l 3]
    set dir  [lindex $l 4]
    if { ($func == "") || [string match {[ ]*} $func] } {
	set func "\{\}"
    }
    puts $ftcl "Rtf::registerVar $func $var $fifo $type $dir"
}


close $ftcl

exit











