/*=================================================
	rt_fifo  Tcl extensions

Description:
This file contains the rt_fifo extensions to Tcl/Tk.

-------------------------
Charles Andr 2000
This code is released as open source.   We hope you find 
it useful.  If you discover a bug or add an enhancement 
we'd like to hear from you. 

	andre@unice.fr
------------------------- 
=================================================*/
/*
 * $Log: rtftclx.c $
 * Revision 1.1 (initial)  2000/03/31 andre
 * Revision 1.2            2000/04/26 andre
 * Revision 1.3            2000/06/26 andre
 */
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <tcl.h>

static int rtfDebug; /* initialized to false (0) */

#define BUF_SIZE 128
int rtf_debugObjCmd(ClientData clientData,
		Tcl_Interp *interp,
		int objc, 
		Tcl_Obj *CONST objv[]);

int rtf_openObjCmd(ClientData clientData,
		Tcl_Interp *interp,
		int objc, 
		Tcl_Obj *CONST objv[]);

int rtf_closeObjCmd(ClientData clientData,
		Tcl_Interp *interp,
		int objc, 
		Tcl_Obj *CONST objv[]);

int rtf_readObjCmd(ClientData clientData,
		Tcl_Interp *interp,
		int objc, 
		Tcl_Obj *CONST objv[]);

int rtf_writeObjCmd(ClientData clientData,
		Tcl_Interp *interp,
		int objc, 
		Tcl_Obj *CONST objv[]);

/* *** auxiliary functions *** */

/* return 1 if illegal num.
   Otherwise returns 0 and sets name to /dev/rtf$num 
*/
static int mkRtfName (Tcl_Interp *interp,int num, char *name) {
 if ( (num < 0) || (num > 63) ) {
   sprintf(interp->result,"Illegal rtf number: %d. Should be between 0 and 63\n",num);
   return 1;
 }
 sprintf(name,"/dev/rtf%d",num);
 return 0;
}


static  void dumpBuffer (unsigned char *buffer,int length ) {
  int i;
  fprintf(stderr,"`0x");
  for (i=0;i<length;i++) fprintf(stderr,"%02x",buffer[i]);
  fprintf(stderr,"'\n");

}


/*=====================================================
	Rtftclx_Init - entry point
=====================================================*/
int Rtftclx_Init(Tcl_Interp *interp)
{
Tcl_CreateObjCommand(interp, "rtf_debug", rtf_debugObjCmd,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
Tcl_CreateObjCommand(interp, "rtf_open", rtf_openObjCmd,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
Tcl_CreateObjCommand(interp, "rtf_close", rtf_closeObjCmd,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
Tcl_CreateObjCommand(interp, "rtf_read", rtf_readObjCmd,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
Tcl_CreateObjCommand(interp, "rtf_write", rtf_writeObjCmd,
	(ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
Tcl_PkgProvide(interp, "Rtftclx", "1.3");
return TCL_OK;
} /* end RtfTclX_Init */


/*===================================================
	rtf_debugObjCmd - entry point

	rtf_debug boolean

===================================================*/
int rtf_debugObjCmd(ClientData clientData, 
	Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
  int tmp;
  if (objc != 2) {
    Tcl_WrongNumArgs(interp,1,objv,"wrong # args: should be \"rtf_debug boolean\"");
    rtfDebug = 0;
    return TCL_ERROR;
  }
  if (Tcl_GetIntFromObj(interp, objv[1], &tmp) != TCL_OK) {
    sprintf(interp->result,"Invalid int:%X\n",tmp);
    rtfDebug = 0;
    return TCL_ERROR;
  } 
  if (tmp) rtfDebug = 1; else rtfDebug = 0;
  return TCL_OK;
}
/* ****   end rtf_debug **** */



/*===================================================
	rtf_openObjCmd - entry point

	rtf_open rtf-num access

	rtf-num : int
	access : char (r | w)

===================================================*/
int rtf_openObjCmd(ClientData clientData, 
	Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
  int num;
  int length;
  char *access;
  char rtf_name[12];
  Tcl_Channel rtf_r;
  int rtf;
  Tcl_Obj *resultPtr;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp,1,objv,"wrong # args: should be \"rtf_open rtf-num access\"");
    return TCL_ERROR;
  }
  if (Tcl_GetIntFromObj(interp, objv[1], &num) != TCL_OK) {
    sprintf(interp->result,"Invalid int:%X\n",num);
    return TCL_ERROR;
  } 
  if (mkRtfName(interp,num,rtf_name)) return TCL_ERROR;

  access = Tcl_GetStringFromObj(objv[2],&length);
  if (length != 1) {
    sprintf(interp->result,"Invalid access: %s.  Should be 'r' or 'w'\n",access);
    return TCL_ERROR;
  }
  
  switch (*access) {
  case 'r' : 
    if ( (rtf = open(rtf_name,O_RDONLY)) < 0) {
      sprintf(interp->result,"Error when opening %s for reading\n",rtf_name);
      return TCL_ERROR;
    } else {
      if (rtfDebug)
	fprintf(stderr,"%s opened for reading. #%d\n",rtf_name,rtf);
      resultPtr = Tcl_GetObjResult(interp);
      Tcl_SetIntObj(resultPtr, rtf);
      return TCL_OK;
    }
    break;
  case 'w' :
    if ( (rtf = open(rtf_name,O_WRONLY)) < 0) {
      sprintf(interp->result,"Error when opening %s for writing\n",rtf_name);
      return TCL_ERROR;
    } else {
      if (rtfDebug)
	fprintf(stderr,"%s opened for writing. #%d\n",rtf_name,rtf);
      resultPtr = Tcl_GetObjResult(interp);
      Tcl_SetIntObj(resultPtr, rtf);
      return TCL_OK;
    }
    break;
  default :
    sprintf(interp->result,"Invalid access: %c. Should be 'r' or 'w'\n",*access);
    return TCL_ERROR;
  }
  
  
} /* end rtf_openObjCmd */

/*===================================================
	rtf_closeObjCmd - entry point

	rtf_close rtf 

	rtf-num : int


===================================================*/
int rtf_closeObjCmd(ClientData clientData, 
	Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
  int rtf;
  Tcl_Obj *resultPtr;
  
  if (objc != 2) {
    Tcl_WrongNumArgs(interp,1,objv,"wrong # args: should be \"rtf_close rtf\"");
    return TCL_ERROR;
  }
  if (Tcl_GetIntFromObj(interp, objv[1], &rtf) != TCL_OK) {
    sprintf(interp->result,"Invalid int:%X\n",rtf);
    return TCL_ERROR;
  } 
  
  if (close(rtf)==-1) {
    fprintf(stderr,"rtf_close: error when closing %d\n",rtf);
    return TCL_ERROR;
  }
  if (rtfDebug)
    fprintf(stderr,"rtf_close %d\n",rtf);
  return TCL_OK;
 
} /* end rtf_closeObjCmd */


/*===================================================
	rtf_readObjCmd - entry point

	rtf_read rtf nb-of-bytes

	rtf : int
	nbOfBytes : int

===================================================*/

int rtf_readObjCmd(ClientData clientData, 
	Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
  int rtf;
  int nbOfBytes;
  int bytesRead;
  unsigned char buffer[BUF_SIZE];
  Tcl_Obj *resultPtr;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp,1,objv,"wrong # args: should be \"rtf_read rtf nb-of-bytes\"");
    return TCL_ERROR;
  }
  Tcl_GetIntFromObj(interp, objv[1], &rtf); 
  Tcl_GetIntFromObj(interp, objv[2], &nbOfBytes); 
  /* fprintf(stderr,"rtf_read %d %d\n",rtf,nbOfBytes); */
  
  if ((bytesRead=read(rtf,buffer,nbOfBytes))==-1) {
    sprintf(interp->result,"Error when reading rtf%d\n",rtf);
    return TCL_ERROR;
  }
  if (rtfDebug) {
    fprintf(stderr,"rtf_read:rtf=%d,nbOfBytes=%d,bytesRead=%d,",rtf,nbOfBytes,bytesRead);
    dumpBuffer(buffer,bytesRead);
  }
  resultPtr = Tcl_GetObjResult(interp);
  Tcl_SetByteArrayObj(resultPtr,buffer,bytesRead);
  return TCL_OK;
}

/* end rtf_read */

/*===================================================
	rtf_writeObjCmd - entry point

	rtf_write rtf byte-array nb-of-bytes

	rtf : int
	byte-array : ByteArray
 

===================================================*/

int rtf_writeObjCmd(ClientData clientData, 
	Tcl_Interp *interp,
	int objc, Tcl_Obj *CONST objv[])
{
  int rtf;
  int written;
  int length;
  unsigned char *buffer;
  Tcl_Obj *resultPtr;
  
  if (objc != 3) {
    Tcl_WrongNumArgs(interp,1,objv,"wrong # args: should be \"rtf_write rtf byte-array\"");
    return TCL_ERROR;
  }
  Tcl_GetIntFromObj(interp, objv[1], &rtf); 
  buffer=Tcl_GetByteArrayFromObj(objv[2], &length); 
  if (rtfDebug) {
    fprintf(stderr,"rtf_write:rtf=%d,length=%d,buffer=",rtf,length);
    dumpBuffer(buffer,length);
  }
  if ((written=write(rtf,buffer,length))==-1) {
    sprintf(interp->result,"Error when writing rtf%d\n",rtf);
    return TCL_ERROR;
  }
  resultPtr = Tcl_GetObjResult(interp);
  Tcl_SetIntObj(resultPtr,written);
  return TCL_OK;
}
/* end rtf_write */









