/* testRtf.c */
/* Charles Andr */
/* June 26, 2000 */


#define MODULE

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>


#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>
#include <math.h>

#include "./testRtfCfg.h"



#define MONITOR_SEND_FIFO 0
#define MONITOR_RECV_FIFO 1

#define PARAMS_FIFO       62
#define TRACE_FIFO        63
#define STACK_SIZE 2000
#define LOOPS 1000000000


/* default values */
int period=1000000000; /* nanosec */
int actual_period;
int resolution=255;
int oversamp=2;
RTIME tick_period;
RTIME now;

static RT_TASK monitor_task;
static RT_TASK collector_task;
static RT_TASK testRtf_task;
static RT_TASK pwm_task;

static SEM monitor_sem;
static SEM params_sem;
static SEM new_data_sem;

static int data = 0;

static void testRtf (int dummy) 
{
  unsigned int loops = LOOPS;
  int hh=0;
  int mm=0;
  int ss=0;
  int ts=0;

  /* 
     putTo(TRACE_FIFO,testRtf,hh);  
     putTo(TRACE_FIFO,testRtf,mm); 
  */ 
  while (loops--) {
    if (++ts == 10) {
      ts = 0;
      if (++ss == 60) {
	ss = 0;
	if (++mm == 60) {
	  mm = 0;
	  /* putTo(TRACE_FIFO,testRtf,mm); */
	  if (++hh == 24) {
	    hh = 0;
	  }
	  /* putTo(TRACE_FIFO,testRtf,hh); */
	}
	/* putTo(TRACE_FIFO,testRtf,mm); */
      }
      putTo(TRACE_FIFO,testRtf,ss);  
    }
    putTo(TRACE_FIFO,testRtf,ts);  
    rt_task_wait_period();
  }
}
/* end testRtf */

static void pwm (int dummy) 
{
  int loops = 1000;
  int t;
  int temp;
  int lit0,lit1;
  RTIME it,it0,it1; /* internal count */
  int refreshed;

  /* initializations */
  t = data;
  t += 128; /* so that 0 <= t <= 255 */
  it = nano2count(t*(actual_period/resolution));
  it0 = tick_period-it;
  it1 = it;
  lit0 = (int) it0;
  lit1 = (int) it1;

  while (loops--) {
    /* repeatitive sampling */
    int i;

    if ( (loops %1) == 0) putTo(TRACE_FIFO,pwm,loops);

    if (rt_sem_wait_if(&new_data_sem) > 0) {
      /* a new data is available: refresh it */
      t = data;
      t += 128; /* so that 0 <= t <= 255 */
      /*
	putTo(TRACE_FIFO,pwm,t); 
      */  
      it = nano2count(t*(actual_period/resolution));
      /*
	caveats: don't exceed size of long with the * 
	for usual frequency use (t*actual_period)/resolution instead 
      */ 
      it0 = tick_period-it;
      it1 = it;
      lit0 = (int) it0;
      lit1 = (int) it1;
      /* 
	 putTo(TRACE_FIFO,pwm,lit0); 
	 putTo(TRACE_FIFO,pwm,lit1); 
      */
      refreshed = 1;
    } else {
      /* let data,t,it,it0,it1 unchanged */
      refreshed = 0;
    }
    putTo(TRACE_FIFO,pwm,refreshed);
    for(i=0;i<oversamp;i++) {
      if(it0) {
	/* 
	   char temp = inb(port);            
	   temp &= 0xfc;
	   outb(temp,port);
	*/
	temp = 0;
	putTo(TRACE_FIFO,pwm,temp);
	rt_task_make_periodic(rt_whoami(),now+=it0,tick_period);
	rt_task_wait_period();
      };
      
      if(it1) {
	/* 
	   char temp = inb(port);
	   temp |= 3;
	   outb(temp,port);
	*/
	temp = 1;
	putTo(TRACE_FIFO,pwm,temp);
	rt_task_make_periodic(rt_whoami(),now+=it1,tick_period);
	rt_task_wait_period();
      }
    }
  }
  
  rt_task_wait_period();
}
/* end pwm */


/* 
   global static variables indirectly assigned by tcl 
   collector MUST be the only function allowed to assign these variables.
   Other tasks are readers.
   => no need for mutex, even if collector is a low priority task.
*/
static void collector(int dummy) {
  int echo;
  int trash;


  while (rt_sem_wait(&params_sem) >= 0) {
    /* 
       there is a parameter to be read. Get it.
       Variable is chosen by the token
    */
    rtf_get(PARAMS_FIFO, &_token, sizeof(int));
    /* wait for the message */
    rt_sem_wait(&params_sem); 
    switch (_token) {
      caseDataFrom(PARAMS_FIFO,,data) 
      rt_sem_signal(&new_data_sem);
      echo = _token;
      break;
    default :
      /* not assignable variable : discard what is in the rtf */
      /* I'have tried : rtf_reset(PARAMS_FIFO); but does not work
         so I flush as if it was an integer (trash variable)
         Try to adopt a safer cleaning. */
      rtf_get(PARAMS_FIFO, &trash, sizeof(trash));
      echo = -_token;
    }
    putTo(TRACE_FIFO,collector,echo);
  }
}



/* 4-state machine:
   0 : idle
   1 : active (runs timer0 and timer1 in alternation)
   2 : frozen (freezes timer0 and timer1)
   3 : final state

   transitions:
   0 -g-> 1       ; Go
   1 -s-> 3       ; Stop
   1 -f-> 2       ; Freeze
   2 -s-> 3       ; Stop
   2 -r-> 1       ; Resume
*/
static void monitor(int dummy) {
  
  unsigned char cmd, rsp;
  int state = 0; /* initially: idle state */


  while (state != 3) {
    /* wait for the keyboard pressed: 
       wait for event monitor_sem;
       if not available then queues up in priority order */
    while (rt_sem_wait(&monitor_sem) >= 0) {
      /* something available: get it 
	 Read a block from MONITOR_SEND_FIFO,
	 command is the buffer,
	 sizeof(command) is the number of bytes to read 
	 
	 Often used in conjunction with rtl_create_handler to process data
	 received asynchronously from a Linux process */
      monitorGet(cmd);

      switch (state) {
      case 0 :
	/* idle state */
	switch (cmd) {
	case 'g' :
	  /* acknowledge */
	  rsp = 'a';
	  /* start the timer with a period in internal count units.
	     nano2count converts the ns in internal counts.
	     tick_period is in internal count too*/
	  actual_period = period/oversamp;
	  tick_period = start_rt_timer((int)nano2count(actual_period));    
	  /* initialize the testRtf task */
	  /* priority:0, no fpu, no signal */
	  rt_task_init(&testRtf_task,testRtf,0,STACK_SIZE,0,0,0);
	  /* priority:1, no fpu, no signal */
	  rt_task_init(&pwm_task,pwm,0,STACK_SIZE,1,0,0);
	  now = rt_get_time();
	  rt_task_make_periodic(&testRtf_task,now+tick_period,tick_period);
	  rt_task_make_periodic(&pwm_task,now+tick_period,tick_period);
	  state = 1;
	  break;
	default :
	  /* unexpected cmd */
	  rsp = '?';
	}
	break;
      case 1:
	/* active state */
	switch (cmd) {
	case 's' :
	  /* acknowledge */
	  rsp = 'q';
	  state = 3;
	  break;
	case 'f' :
	  rsp = 'a';
	  rt_task_suspend(&pwm_task);
	  state = 2;
	  break;
	default :
	  /* unexpected cmd */
	  rsp = '?';
	}
	break;
      case 2 :
	/* frozen state */
	switch (cmd) {
	case 's' :
	  /* acknowledge */
	  rsp = 'q';
	  state = 3;
	  break;
	case 'r' :
	  /* acknowledge */
	  rsp = 'a';
	  now = rt_get_time();
	  rt_task_resume(&pwm_task);
	  state = 1;
	  break;
	default :
	  /* unexpected cmd */
	  rsp = '?';
	}
	break;
      }
      if (state == 3) {
	stop_rt_timer();
	/* send quit to the operator */
      }
      /* send response */
      monitorPut(rsp);
    }
  }
  /* in state 3 */
}

 
/* handler called whenever a keyborad button is pressed */
static int fifo_handler(unsigned int fifo)
{
  rt_sem_signal(&monitor_sem);
  return 0;
}

/* handler called whenever a parameter is sent */
static int fifo_params_handler(unsigned int fifo)
{
  rt_sem_signal(&params_sem);
  return 0;
}

int init_module(void) {
  rt_set_periodic_mode();
 
  /* creation of fifos for the communication with Linux processes */
  rtf_create(MONITOR_SEND_FIFO,4000);
  rtf_create(MONITOR_RECV_FIFO,4000);
  rtf_create(PARAMS_FIFO,4000);
  rtf_create(TRACE_FIFO,4000);

  /* creation of a handler associated with MONITOR_SEND_FIFO */
  rtf_create_handler(MONITOR_SEND_FIFO,fifo_handler);
  rtf_create_handler(PARAMS_FIFO,fifo_params_handler);

  /* creation of semaphores used for intertask communication */
  rt_sem_init(&monitor_sem,0);
  rt_sem_init(&params_sem,0);
  rt_sem_init(&new_data_sem,0);

  /* creation of real-time service tasks */
  /* priority:2, no fpu, no signal */
  rt_task_init(&monitor_task,monitor,0,4000,2,0,0);
  rt_task_init(&collector_task,collector,0,4000,2,0,0);
 
  /* execute service tasks */
  rt_task_resume(&monitor_task);
  rt_task_resume(&collector_task);
  return 0;
}


void cleanup_module(void) {

  /* 
  stop_rt_timer();
  */
  /* delete rt-fifos */
  rtf_destroy(MONITOR_RECV_FIFO);
  rtf_destroy(TRACE_FIFO);
  rtf_destroy(MONITOR_SEND_FIFO);
  rtf_destroy(PARAMS_FIFO);
  /* delete tasks */
  rt_task_delete(&pwm_task);
  rt_task_delete(&testRtf_task);
  rt_task_delete(&monitor_task);
  rt_task_delete(&collector_task);
}




